from pathlib import Path
from traceback import print_exc
from typing import Literal, Union, Text, Optional, Tuple

import matplotlib.pyplot as plt
from pyannote.core import Annotation, Segment, SlidingWindowFeature, notebook
from pyannote.database.util import load_rttm
from pyannote.metrics.diarization import DiarizationErrorRate
from rx.core import Observer


class RTTMWriter(Observer):
    def __init__(self, path: Union[Path, Text], patch_collar: float = 0.05):
        super().__init__()
        self.patch_collar = patch_collar
        self.path = Path(path)
        if self.path.exists():
            self.path.unlink()

    def patch_rttm(self):
        """Stitch same-speaker turns that are close to each other"""
        annotation = list(load_rttm(self.path).values())[0]
        with open(self.path, 'w') as file:
            annotation.support(self.patch_collar).write_rttm(file)

    def on_next(self, value: Tuple[Annotation, Optional[SlidingWindowFeature]]):
        with open(self.path, 'a') as file:
            value[0].write_rttm(file)

    def on_error(self, error: Exception):
        try:
            self.patch_rttm()
        except Exception:
            print("Error while patching RTTM file:")
            print_exc()
            exit(1)

    def on_completed(self):
        self.patch_rttm()


class RealTimePlot(Observer):
    def __init__(
        self,
        duration: float,
        latency: float,
        visualization: Literal["slide", "accumulate"] = "slide",
        reference: Optional[Union[Path, Text]] = None,
    ):
        super().__init__()
        assert visualization in ["slide", "accumulate"]
        self.visualization = visualization
        self.reference = reference
        if self.reference is not None:
            self.reference = list(load_rttm(reference).values())[0]
        self.window_duration = duration
        self.latency = latency
        self.figure, self.axs, self.num_axs = None, None, -1

    def _init_num_axs(self, waveform: Optional[SlidingWindowFeature]):
        if self.num_axs == -1:
            self.num_axs = 1
            if waveform is not None:
                self.num_axs += 1
            if self.reference is not None:
                self.num_axs += 1

    def _init_figure(self, waveform: Optional[SlidingWindowFeature]):
        self._init_num_axs(waveform)
        self.figure, self.axs = plt.subplots(self.num_axs, 1, figsize=(10, 2 * self.num_axs))
        if self.num_axs == 1:
            self.axs = [self.axs]

    def _clear_axs(self):
        for i in range(self.num_axs):
            self.axs[i].clear()

    def get_plot_bounds(self, real_time: float) -> Segment:
        start_time = 0
        end_time = real_time - self.latency
        if self.visualization == "slide":
            start_time = max(0., end_time - self.window_duration)
        return Segment(start_time, end_time)

    def on_next(self, values: Tuple[Annotation, SlidingWindowFeature, float]):
        prediction, waveform, real_time = values
        # Initialize figure if first call
        if self.figure is None:
            self._init_figure(waveform)
        # Clear previous plots
        self._clear_axs()
        # Set plot bounds
        notebook.crop = self.get_plot_bounds(real_time)

        # Plot current values
        if self.reference is not None:
            metric = DiarizationErrorRate()
            mapping = metric.optimal_mapping(self.reference, prediction)
            prediction.rename_labels(mapping=mapping, copy=False)
        notebook.plot_annotation(prediction, self.axs[0])
        self.axs[0].set_title("Output")
        if self.num_axs == 2:
            if waveform is not None:
                notebook.plot_feature(waveform, self.axs[1])
                self.axs[1].set_title("Audio")
            elif self.reference is not None:
                notebook.plot_annotation(self.reference, self.axs[1])
                self.axs[1].set_title("Reference")
        elif self.num_axs == 3:
            notebook.plot_feature(waveform, self.axs[1])
            self.axs[1].set_title("Audio")
            notebook.plot_annotation(self.reference, self.axs[2])
            self.axs[2].set_title("Reference")

        # Draw
        plt.tight_layout()
        self.figure.canvas.draw()
        self.figure.canvas.flush_events()
        plt.pause(0.05)

    def on_error(self, error: Exception):
        print_exc()
        exit(1)



from dataclasses import dataclass
from typing import Callable, Optional, List, Any, Tuple

import numpy as np
import rx
from pyannote.core import Annotation, SlidingWindow, SlidingWindowFeature, Segment
from rx import operators as ops
from rx.core import Observable

Operator = Callable[[Observable], Observable]


@dataclass
class AudioBufferState:
    chunk: Optional[np.ndarray]
    buffer: Optional[np.ndarray]
    start_time: float
    changed: bool

    @staticmethod
    def initial():
        return AudioBufferState(None, None, 0, False)

    @staticmethod
    def has_samples(num_samples: int):
        def call_fn(state) -> bool:
            return state.chunk is not None and state.chunk.shape[1] == num_samples
        return call_fn

    @staticmethod
    def to_sliding_window(sample_rate: int):
        def call_fn(state) -> SlidingWindowFeature:
            resolution = SlidingWindow(start=state.start_time, duration=1. / sample_rate, step=1. / sample_rate)
            return SlidingWindowFeature(state.chunk.T, resolution)
        return call_fn


def regularize_stream(
    duration: float = 5,
    step: float = 0.5,
    sample_rate: int = 16000
) -> Operator:
    chunk_samples = int(round(sample_rate * duration))
    step_samples = int(round(sample_rate * step))

    def accumulate(state: AudioBufferState, value: np.ndarray):
        # State contains the last emitted chunk, the current step buffer and the last starting time
        if value.ndim != 2 or value.shape[0] != 1:
            raise ValueError(f"Waveform must have shape (1, samples) but {value.shape} was found")
        start_time = state.start_time

        # Add new samples to the buffer
        buffer = value if state.buffer is None else np.concatenate([state.buffer, value], axis=1)

        # Check for buffer overflow
        if buffer.shape[1] >= step_samples:
            # Pop samples from buffer
            if buffer.shape[1] == step_samples:
                new_chunk, new_buffer = buffer, None
            else:
                new_chunk = buffer[:, :step_samples]
                new_buffer = buffer[:, step_samples:]

            # Add samples to next chunk
            if state.chunk is not None:
                new_chunk = np.concatenate([state.chunk, new_chunk], axis=1)

            # Truncate chunk to ensure a fixed duration
            if new_chunk.shape[1] > chunk_samples:
                new_chunk = new_chunk[:, -chunk_samples:]
                start_time += step

            # Chunk has changed because of buffer overflow
            return AudioBufferState(new_chunk, new_buffer, start_time, changed=True)

        # Chunk has not changed
        return AudioBufferState(state.chunk, buffer, start_time, changed=False)

    return rx.pipe(
        # Accumulate last <=duration seconds of waveform as an AudioBufferState
        ops.scan(accumulate, AudioBufferState.initial()),
        # Take only states that have the desired duration and whose chunk has changed since last time
        ops.filter(AudioBufferState.has_samples(chunk_samples)),
        ops.filter(lambda state: state.changed),
        # Transform state into a SlidingWindowFeature containing the new chunk
        ops.map(AudioBufferState.to_sliding_window(sample_rate))
    )


def buffer_slide(n: int):
    def accumulate(state: List[Any], value: Any) -> List[Any]:
        new_state = [*state, value]
        if len(new_state) > n:
            return new_state[1:]
        return new_state
    return rx.pipe(ops.scan(accumulate, []))


@dataclass
class PredictionWithAudio:
    prediction: Annotation
    waveform: Optional[SlidingWindowFeature] = None

    @property
    def has_audio(self) -> bool:
        return self.waveform is not None


@dataclass
class OutputAccumulationState:
    annotation: Optional[Annotation]
    waveform: Optional[SlidingWindowFeature]
    real_time: float
    next_sample: Optional[int]

    @staticmethod
    def initial() -> 'OutputAccumulationState':
        return OutputAccumulationState(None, None, 0, 0)

    @property
    def cropped_waveform(self) -> SlidingWindowFeature:
        return SlidingWindowFeature(
            self.waveform[:self.next_sample],
            self.waveform.sliding_window,
        )

    def to_tuple(self) -> Tuple[Optional[Annotation], Optional[SlidingWindowFeature], float]:
        return self.annotation, self.cropped_waveform, self.real_time


def accumulate_output(
    duration: float,
    step: float,
    patch_collar: float = 0.05,
) -> Operator:
    """Accumulate predictions and audio to infinity: O(N) space complexity.
    Uses a pre-allocated buffer that doubles its size once full: O(logN) concat operations.

    Parameters
    ----------
    duration: float
        Buffer duration in seconds.
    step: float
        Duration of the chunks at each event in seconds.
        The first chunk may be bigger given the latency.
    patch_collar: float, optional
        Collar to merge speaker turns of the same speaker, in seconds.
        Defaults to 0.05 (i.e. 50ms).
    Returns
    -------
    A reactive x operator implementing this behavior.
    """
    def accumulate(
        state: OutputAccumulationState,
        value: Tuple[Annotation, Optional[SlidingWindowFeature]]
    ) -> OutputAccumulationState:
        value = PredictionWithAudio(*value)
        annotation, waveform = None, None

        # Determine the real time of the stream
        real_time = duration if state.annotation is None else state.real_time + step

        # Update total annotation with current predictions
        if state.annotation is None:
            annotation = value.prediction
        else:
            annotation = state.annotation.update(value.prediction).support(patch_collar)

        # Update total waveform if there's audio in the input
        new_next_sample = 0
        if value.has_audio:
            num_new_samples = value.waveform.data.shape[0]
            new_next_sample = state.next_sample + num_new_samples
            sw_holder = state
            if state.waveform is None:
                # Initialize the audio buffer with 10 times the size of the first chunk
                waveform, sw_holder = np.zeros((10 * num_new_samples, 1)), value
            elif new_next_sample < state.waveform.data.shape[0]:
                # The buffer still has enough space to accommodate the chunk
                waveform = state.waveform.data
            else:
                # The buffer is full, double its size
                waveform = np.concatenate(
                    (state.waveform.data, np.zeros_like(state.waveform.data)), axis=0
                )
            # Copy chunk into buffer
            waveform[state.next_sample:new_next_sample] = value.waveform.data
            waveform = SlidingWindowFeature(waveform, sw_holder.waveform.sliding_window)

        return OutputAccumulationState(annotation, waveform, real_time, new_next_sample)

    return rx.pipe(
        ops.scan(accumulate, OutputAccumulationState.initial()),
        ops.map(OutputAccumulationState.to_tuple),
    )


def buffer_output(
    duration: float,
    step: float,
    latency: float,
    sample_rate: int,
    patch_collar: float = 0.05,
) -> Operator:
    """Store last predictions and audio inside a fixed buffer.
    Provides the best time/space complexity trade-off if the past data is not needed.

    Parameters
    ----------
    duration: float
        Buffer duration in seconds.
    step: float
        Duration of the chunks at each event in seconds.
        The first chunk may be bigger given the latency.
    latency: float
        Latency of the system in seconds.
    sample_rate: int
        Sample rate of the audio source.
    patch_collar: float, optional
        Collar to merge speaker turns of the same speaker, in seconds.
        Defaults to 0.05 (i.e. 50ms).

    Returns
    -------
    A reactive x operator implementing this behavior.
    """
    # Define some useful constants
    num_samples = int(round(duration * sample_rate))
    num_step_samples = int(round(step * sample_rate))
    resolution = 1 / sample_rate

    def accumulate(
        state: OutputAccumulationState,
        value: Tuple[Annotation, Optional[SlidingWindowFeature]]
    ) -> OutputAccumulationState:
        value = PredictionWithAudio(*value)
        annotation, waveform = None, None

        # Determine the real time of the stream and the start time of the buffer
        real_time = duration if state.annotation is None else state.real_time + step
        start_time = max(0., real_time - latency - duration)

        # Update annotation and constrain its bounds to the buffer
        if state.annotation is None:
            annotation = value.prediction
        else:
            annotation = state.annotation.update(value.prediction) \
                .support(patch_collar) \
                .extrude(Segment(0, start_time))

        # Update the audio buffer if there's audio in the input
        new_next_sample = state.next_sample + num_step_samples
        if value.has_audio:
            if state.waveform is None:
                # Determine the size of the first chunk
                expected_duration = duration + step - latency
                expected_samples = int(round(expected_duration * sample_rate))
                # Shift indicator to start copying new audio in the buffer
                new_next_sample = state.next_sample + expected_samples
                # Buffer size is duration + step
                waveform = np.zeros((num_samples + num_step_samples, 1))
                # Copy first chunk into buffer (slicing because of rounding errors)
                waveform[:expected_samples] = value.waveform.data[:expected_samples]
            elif state.next_sample <= num_samples:
                # The buffer isn't full, copy into next free buffer chunk
                waveform = state.waveform.data
                waveform[state.next_sample:new_next_sample] = value.waveform.data
            else:
                # The buffer is full, shift values to the left and copy into last buffer chunk
                waveform = np.roll(state.waveform.data, -num_step_samples, axis=0)
                waveform[-num_step_samples:] = value.waveform.data

            # Wrap waveform in a sliding window feature to include timestamps
            window = SlidingWindow(start=start_time, duration=resolution, step=resolution)
            waveform = SlidingWindowFeature(waveform, window)

        return OutputAccumulationState(annotation, waveform, real_time, new_next_sample)

    return rx.pipe(
        ops.scan(accumulate, OutputAccumulationState.initial()),
        ops.map(OutputAccumulationState.to_tuple),
    )



import argparse
from pathlib import Path

import diart.operators as dops
import diart.sources as src
import rx.operators as ops
from diart.pipelines import OnlineSpeakerDiarization
from diart.sinks import RealTimePlot, RTTMWriter

# Define script arguments
parser = argparse.ArgumentParser()
parser.add_argument("source", type=str, help="Path to an audio file | 'microphone'")
parser.add_argument("--step", default=0.5, type=float, help="Source sliding window step")
parser.add_argument("--latency", default=0.5, type=float, help="System latency")
parser.add_argument("--sample-rate", default=16000, type=int, help="Source sample rate")
parser.add_argument("--tau", default=0.5, type=float, help="Activity threshold tau active")
parser.add_argument("--rho", default=0.3, type=float, help="Speech duration threshold rho update")
parser.add_argument("--delta", default=1, type=float, help="Maximum distance threshold delta new")
parser.add_argument("--gamma", default=3, type=float, help="Parameter gamma for overlapped speech penalty")
parser.add_argument("--beta", default=10, type=float, help="Parameter beta for overlapped speech penalty")
parser.add_argument("--max-speakers", default=20, type=int, help="Maximum number of identifiable speakers")
parser.add_argument(
    "--output", type=str,
    help="Output directory to store the RTTM. Defaults to home directory "
         "if source is microphone or parent directory if source is a file"
)
args = parser.parse_args()

# Define online speaker diarization pipeline
pipeline = OnlineSpeakerDiarization(
    step=args.step,
    latency=args.latency,
    tau_active=args.tau,
    rho_update=args.rho,
    delta_new=args.delta,
    gamma=args.gamma,
    beta=args.beta,
    max_speakers=args.max_speakers,
)

# Manage audio source
if args.source != "microphone":
    args.source = Path(args.source).expanduser()
    uri = args.source.name.split(".")[0]
    output_dir = args.source.parent if args.output is None else Path(args.output)
    audio_source = src.FileAudioSource(
        file=args.source,
        uri=uri,
        reader=src.RegularAudioFileReader(
            args.sample_rate, pipeline.duration, pipeline.step
        ),
    )
else:
    output_dir = Path("~/").expanduser() if args.output is None else Path(args.output)
    audio_source = src.MicrophoneAudioSource(args.sample_rate)

# Build pipeline from audio source and stream predictions to a real-time plot
pipeline.from_source(audio_source).pipe(
    ops.do(RTTMWriter(path=output_dir / "output.rttm")),
    dops.buffer_output(
        duration=pipeline.duration,
        step=pipeline.step,
        latency=pipeline.latency,
        sample_rate=audio_source.sample_rate
    ),
).subscribe(RealTimePlot(pipeline.duration, pipeline.latency))

# Read audio source as a stream
if args.source == "microphone":
    print("Recording...")
audio_source.read()



from typing import Union, Optional, List, Literal, Iterable, Tuple

import numpy as np
import torch
from pyannote.audio.pipelines.utils import PipelineModel, get_model, get_devices
from pyannote.audio.utils.signal import Binarize as PyanBinarize
from pyannote.core import Annotation, Segment, SlidingWindow, SlidingWindowFeature

from .mapping import SpeakerMap, SpeakerMapBuilder


class FrameWiseModel:
    def __init__(self, model: PipelineModel, device: Optional[torch.device] = None):
        self.model = get_model(model)
        self.model.eval()
        if device is None:
            device = get_devices(needs=1)[0]
        self.model.to(device)

    def __call__(self, waveform: SlidingWindowFeature) -> SlidingWindowFeature:
        with torch.no_grad():
            wave = torch.from_numpy(waveform.data.T[np.newaxis])
            output = self.model(wave.to(self.model.device)).cpu().numpy()[0]
        # Temporal resolution of the output
        resolution = self.model.specifications.duration / output.shape[0]
        # Temporal shift to keep track of current start time
        resolution = SlidingWindow(
            start=waveform.sliding_window.start,
            duration=resolution,
            step=resolution
        )
        return SlidingWindowFeature(output, resolution)


class ChunkWiseModel:
    def __init__(self, model: PipelineModel, device: Optional[torch.device] = None):
        self.model = get_model(model)
        self.model.eval()
        if device is None:
            device = get_devices(needs=1)[0]
        self.model.to(device)

    def __call__(self, waveform: SlidingWindowFeature, weights: Optional[SlidingWindowFeature]) -> torch.Tensor:
        with torch.no_grad():
            chunk = torch.from_numpy(waveform.data.T).float()
            inputs = chunk.unsqueeze(0).to(self.model.device)
            if weights is not None:
                # weights has shape (num_local_speakers, num_frames)
                weights = torch.from_numpy(weights.data.T).float().to(self.model.device)
                inputs = inputs.repeat(weights.shape[0], 1, 1)
            # Shape (num_speakers, emb_dimension)
            output = self.model(inputs, weights=weights).cpu()
        return output


class OverlappedSpeechPenalty:
    """
    Parameters
    ----------
    gamma: float, optional
        Exponent to lower low-confidence predictions.
        Defaults to 3.
    beta: float, optional
        Softmax's temperature parameter (actually 1/beta) to lower joint speaker activations.
        Defaults to 10.
    """
    def __init__(self, gamma: float = 3, beta: float = 10):
        self.gamma = gamma
        self.beta = beta

    def __call__(self, segmentation: SlidingWindowFeature) -> SlidingWindowFeature:
        weights = torch.from_numpy(segmentation.data).float().T
        with torch.no_grad():
            probs = torch.softmax(self.beta * weights, dim=0)
            weights = torch.pow(weights, self.gamma) * torch.pow(probs, self.gamma)
            weights[weights < 1e-8] = 1e-8
        return SlidingWindowFeature(weights.T.numpy(), segmentation.sliding_window)


class EmbeddingNormalization:
    def __init__(self, norm: Union[float, torch.Tensor] = 1):
        self.norm = norm

    def __call__(self, embeddings: torch.Tensor) -> torch.Tensor:
        if isinstance(self.norm, torch.Tensor):
            assert self.norm.shape[0] == embeddings.shape[0]
        with torch.no_grad():
            norm_embs = self.norm * embeddings / torch.norm(embeddings, p=2, dim=1, keepdim=True)
        return norm_embs


class AggregationStrategy:
    """Abstract class representing a strategy to aggregate overlapping buffers"""

    @staticmethod
    def build(name: Literal["mean", "hamming", "first"]) -> 'AggregationStrategy':
        """Build an AggregationStrategy instance based on its name"""
        assert name in ("mean", "hamming", "first")
        if name == "mean":
            return AverageStrategy()
        elif name == "hamming":
            return HammingWeightedAverageStrategy()
        else:
            return FirstOnlyStrategy()

    def __call__(self, buffers: List[SlidingWindowFeature], focus: Segment) -> SlidingWindowFeature:
        """Aggregate chunks over a specific region.

        Parameters
        ----------
        buffers: list of SlidingWindowFeature, shapes (frames, speakers)
            Buffers to aggregate
        focus: Segment
            Region to aggregate that is shared among the buffers

        Returns
        -------
        aggregation: SlidingWindowFeature, shape (cropped_frames, speakers)
            Aggregated values over the focus region
        """
        aggregation = self.aggregate(buffers, focus)
        resolution = focus.duration / aggregation.shape[0]
        resolution = SlidingWindow(
            start=focus.start,
            duration=resolution,
            step=resolution
        )
        return SlidingWindowFeature(aggregation, resolution)

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        raise NotImplementedError


class HammingWeightedAverageStrategy(AggregationStrategy):
    """Compute the average weighted by the corresponding Hamming-window aligned to each buffer"""

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        num_frames, num_speakers = buffers[0].data.shape
        hamming, intersection = [], []
        for buffer in buffers:
            # Crop buffer to focus region
            b = buffer.crop(focus, fixed=focus.duration)
            # Crop Hamming window to focus region
            h = np.expand_dims(np.hamming(num_frames), axis=-1)
            h = SlidingWindowFeature(h, buffer.sliding_window)
            h = h.crop(focus, fixed=focus.duration)
            hamming.append(h.data)
            intersection.append(b.data)
        hamming, intersection = np.stack(hamming), np.stack(intersection)
        # Calculate weighted mean
        return np.sum(hamming * intersection, axis=0) / np.sum(hamming, axis=0)


class AverageStrategy(AggregationStrategy):
    """Compute a simple average over the focus region"""

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        # Stack all overlapping regions
        intersection = np.stack([
            buffer.crop(focus, fixed=focus.duration)
            for buffer in buffers
        ])
        return np.mean(intersection, axis=0)


class FirstOnlyStrategy(AggregationStrategy):
    """Instead of aggregating, keep the first focus region in the buffer list"""

    def aggregate(self, buffers: List[SlidingWindowFeature], focus: Segment) -> np.ndarray:
        return buffers[0].crop(focus, fixed=focus.duration)


class DelayedAggregation:
    """Aggregate aligned overlapping windows of the same duration
    across sliding buffers with a specific step and latency.

    Parameters
    ----------
    step: float
        Shift between two consecutive buffers, in seconds.
    latency: float, optional
        Desired latency, in seconds. Defaults to step.
        The higher the latency, the more overlapping windows to aggregate.
    strategy: ("mean", "hamming", "any"), optional
        Specifies how to aggregate overlapping windows. Defaults to "hamming".
        "mean": simple average
        "hamming": average weighted by the Hamming window values (aligned to the buffer)
        "any": no aggregation, pick the first overlapping window
    stream_end: float, optional
        Stream end time (in seconds). Defaults to None.
        If the stream end time is known, then append remaining outputs at the end,
        otherwise the last `latency - step` seconds are ignored.

    Example
    --------
    >>> duration = 5
    >>> frames = 500
    >>> step = 0.5
    >>> speakers = 2
    >>> start_time = 10
    >>> resolution = duration / frames
    >>> dagg = DelayedAggregation(step=step, latency=2, strategy="mean")
    >>> buffers = [
    >>>     SlidingWindowFeature(
    >>>         np.random.rand(frames, speakers),
    >>>         SlidingWindow(start=(i + start_time) * step, duration=resolution, step=resolution)
    >>>     )
    >>>     for i in range(dagg.num_overlapping_windows)
    >>> ]
    >>> dagg.num_overlapping_windows
    ... 4
    >>> dagg(buffers).data.shape
    ... (51, 2)  # Rounding errors are possible when cropping the buffers
    """

    def __init__(
        self,
        step: float,
        latency: Optional[float] = None,
        strategy: Literal["mean", "hamming", "first"] = "hamming",
        stream_end: Optional[float] = None
    ):
        self.step = step
        self.latency = latency
        self.strategy = strategy
        self.stream_end = stream_end

        if self.latency is None:
            self.latency = self.step

        assert self.step <= self.latency, "Invalid latency requested"

        self.num_overlapping_windows = int(round(self.latency / self.step))
        self.aggregate = AggregationStrategy.build(self.strategy)

    def _prepend_or_append(
        self,
        output_window: SlidingWindowFeature,
        output_region: Segment,
        buffers: List[SlidingWindowFeature]
    ):
        last_buffer = buffers[-1].extent
        # Prepend prediction until we match the latency in case of first buffer
        if len(buffers) == 1 and last_buffer.start == 0:
            num_frames = output_window.data.shape[0]
            first_region = Segment(0, output_region.end)
            first_output = buffers[0].crop(
                first_region, fixed=first_region.duration
            )
            first_output[-num_frames:] = output_window.data
            resolution = output_region.end / first_output.shape[0]
            output_window = SlidingWindowFeature(
                first_output,
                SlidingWindow(start=0, duration=resolution, step=resolution)
            )
        # Append rest of the outputs
        elif self.stream_end is not None and last_buffer.end == self.stream_end:
            num_frames = output_window.data.shape[0]
            last_region = Segment(output_region.start, last_buffer.end)
            last_output = buffers[-1].crop(
                last_region, fixed=last_region.duration
            )
            last_output[:num_frames] = output_window.data
            resolution = self.latency / last_output.shape[0]
            output_window = SlidingWindowFeature(
                last_output,
                SlidingWindow(
                    start=output_region.start,
                    duration=resolution,
                    step=resolution
                )
            )
        return output_window

    def __call__(self, buffers: List[SlidingWindowFeature]) -> SlidingWindowFeature:
        # Determine overlapping region to aggregate
        start = buffers[-1].extent.end - self.latency
        region = Segment(start, start + self.step)
        return self._prepend_or_append(self.aggregate(buffers, region), region, buffers)


class OnlineSpeakerClustering:
    def __init__(
        self,
        tau_active: float,
        rho_update: float,
        delta_new: float,
        metric: Optional[str] = "cosine",
        max_speakers: int = 20
    ):
        self.tau_active = tau_active
        self.rho_update = rho_update
        self.delta_new = delta_new
        self.metric = metric
        self.max_speakers = max_speakers
        self.centers: Optional[np.ndarray] = None
        self.active_centers = set()
        self.blocked_centers = set()

    @property
    def num_free_centers(self) -> int:
        return self.max_speakers - self.num_known_speakers - self.num_blocked_speakers

    @property
    def num_known_speakers(self) -> int:
        return len(self.active_centers)

    @property
    def num_blocked_speakers(self) -> int:
        return len(self.blocked_centers)

    @property
    def inactive_centers(self) -> List[int]:
        return [
            c
            for c in range(self.max_speakers)
            if c not in self.active_centers or c in self.blocked_centers
        ]

    def get_next_center_position(self) -> Optional[int]:
        for center in range(self.max_speakers):
            if center not in self.active_centers and center not in self.blocked_centers:
                return center

    def init_centers(self, dimension: int):
        self.centers = np.zeros((self.max_speakers, dimension))
        self.active_centers = set()
        self.blocked_centers = set()

    def update(self, assignments: Iterable[Tuple[int, int]], embeddings: np.ndarray):
        if self.centers is not None:
            for l_spk, g_spk in assignments:
                assert g_spk in self.active_centers, "Cannot update unknown centers"
                self.centers[g_spk] += embeddings[l_spk]

    def add_center(self, embedding: np.ndarray) -> int:
        center = self.get_next_center_position()
        self.centers[center] = embedding
        self.active_centers.add(center)
        return center

    def identify(
        self,
        segmentation: SlidingWindowFeature,
        embeddings: torch.Tensor
    ) -> SpeakerMap:
        embeddings = embeddings.detach().cpu().numpy()
        active_speakers = np.where(np.max(segmentation.data, axis=0) >= self.tau_active)[0]
        long_speakers = np.where(np.mean(segmentation.data, axis=0) >= self.rho_update)[0]
        num_local_speakers = segmentation.data.shape[1]

        if self.centers is None:
            self.init_centers(embeddings.shape[1])
            assignments = [
                (spk, self.add_center(embeddings[spk]))
                for spk in active_speakers
            ]
            return SpeakerMapBuilder.hard_map(
                shape=(num_local_speakers, self.max_speakers),
                assignments=assignments,
                maximize=False,
            )

        # Obtain a mapping based on distances between embeddings and centers
        dist_map = SpeakerMapBuilder.dist(embeddings, self.centers, self.metric)
        # Remove any assignments containing invalid speakers
        inactive_speakers = np.array([
            spk for spk in range(num_local_speakers)
            if spk not in active_speakers
        ])
        dist_map = dist_map.unmap_speakers(inactive_speakers, self.inactive_centers)
        # Keep assignments under the distance threshold
        valid_map = dist_map.unmap_threshold(self.delta_new)

        # Some speakers might be unidentified
        missed_speakers = [
            s for s in active_speakers
            if not valid_map.is_source_speaker_mapped(s)
        ]

        # Add assignments to new centers if possible
        new_center_speakers = []
        for spk in missed_speakers:
            has_space = len(new_center_speakers) < self.num_free_centers
            if has_space and spk in long_speakers:
                # Flag as a new center
                new_center_speakers.append(spk)
            else:
                # Cannot create a new center
                # Get global speakers in order of preference
                preferences = np.argsort(dist_map.mapping_matrix[spk, :])
                preferences = [
                    g_spk for g_spk in preferences if g_spk in self.active_centers
                ]
                # Get the free global speakers among the preferences
                _, g_assigned = valid_map.valid_assignments()
                free = [g_spk for g_spk in preferences if g_spk not in g_assigned]
                if free:
                    # The best global speaker is the closest free one
                    valid_map = valid_map.set_source_speaker(spk, free[0])

        # Update known centers
        to_update = [
            (ls, gs)
            for ls, gs in zip(*valid_map.valid_assignments())
            if ls not in missed_speakers and ls in long_speakers
        ]
        self.update(to_update, embeddings)

        # Add new centers
        for spk in new_center_speakers:
            valid_map = valid_map.set_source_speaker(
                spk, self.add_center(embeddings[spk])
            )

        return valid_map

    def __call__(self, segmentation: SlidingWindowFeature, embeddings: torch.Tensor) -> SlidingWindowFeature:
        return SlidingWindowFeature(
            self.identify(segmentation, embeddings).apply(segmentation.data),
            segmentation.sliding_window
        )


class Binarize:
    def __init__(self, uri: str, tau_active: float):
        self.uri = uri
        self._binarize = PyanBinarize(
            onset=tau_active,
            offset=tau_active,
            min_duration_on=0,
            min_duration_off=0,
        )

    def _select(
        self, scores: SlidingWindowFeature, speaker: int
    ) -> SlidingWindowFeature:
        return SlidingWindowFeature(
            scores[:, speaker].reshape(-1, 1), scores.sliding_window
        )

    def __call__(self, segmentation: SlidingWindowFeature) -> Annotation:
        annotation = Annotation(uri=self.uri, modality="speech")
        for speaker in range(segmentation.data.shape[1]):
            turns = self._binarize(self._select(segmentation, speaker))
            for speaker_turn in turns.itersegments():
                annotation[speaker_turn, speaker] = f"speaker{speaker}"
        return annotation



from typing import Optional

import matplotlib.pyplot as plt
from pyannote.core import Annotation, Segment, SlidingWindowFeature, notebook


def visualize_feature(duration: Optional[float] = None):
    def apply(feature: SlidingWindowFeature):
        if duration is None:
            notebook.crop = feature.extent
        else:
            notebook.crop = Segment(feature.extent.end - duration, feature.extent.end)
        plt.rcParams["figure.figsize"] = (8, 2)
        notebook.plot_feature(feature)
        plt.tight_layout()
        plt.show()
    return apply


def visualize_annotation(duration: Optional[float] = None):
    def apply(annotation: Annotation):
        extent = annotation.get_timeline().extent()
        if duration is None:
            notebook.crop = extent
        else:
            notebook.crop = Segment(extent.end - duration, extent.end)
        plt.rcParams["figure.figsize"] = (8, 2)
        notebook.plot_annotation(annotation)
        plt.tight_layout()
        plt.show()
    return apply



from __future__ import annotations

from typing import Callable, Iterable, List, Optional, Text, Tuple, Union

import numpy as np
from pyannote.core.utils.distance import cdist
from scipy.optimize import linear_sum_assignment


class MappingMatrixObjective:
    def invalid_tensor(self, shape: Union[Tuple, int]) -> np.ndarray:
        return np.ones(shape) * self.invalid_value

    def optimal_assignments(self, matrix: np.ndarray) -> List[int]:
        return list(linear_sum_assignment(matrix, self.maximize)[1])

    def mapped_indices(self, matrix: np.ndarray, axis: int) -> List[int]:
        # Entries full of invalid_value are not mapped
        best_values = self.best_value_fn(matrix, axis=axis)
        return list(np.where(best_values != self.invalid_value)[0])

    def hard_speaker_map(
        self, num_src: int, num_tgt: int, assignments: Iterable[Tuple[int, int]]
    ) -> SpeakerMap:
        mapping_matrix = self.invalid_tensor(shape=(num_src, num_tgt))
        for src, tgt in assignments:
            mapping_matrix[src, tgt] = self.best_possible_value
        return SpeakerMap(mapping_matrix, self)

    @property
    def invalid_value(self) -> float:
        # linear_sum_assignment cannot deal with np.inf,
        # which would be ideal. Using a big number instead.
        return -1e10 if self.maximize else 1e10

    @property
    def maximize(self) -> bool:
        raise NotImplementedError()

    @property
    def best_possible_value(self) -> float:
        raise NotImplementedError()

    @property
    def best_value_fn(self) -> Callable:
        raise NotImplementedError()


class MinimizationObjective(MappingMatrixObjective):
    @property
    def maximize(self) -> bool:
        return False

    @property
    def best_possible_value(self) -> float:
        return 0

    @property
    def best_value_fn(self) -> Callable:
        return np.min


class MaximizationObjective(MappingMatrixObjective):
    def __init__(self, max_value: float = 1):
        self.max_value = max_value

    @property
    def maximize(self) -> bool:
        return True

    @property
    def best_possible_value(self) -> float:
        return self.max_value

    @property
    def best_value_fn(self) -> Callable:
        return np.max


class SpeakerMapBuilder:
    @staticmethod
    def hard_map(
        shape: Tuple[int, int], assignments: Iterable[Tuple[int, int]], maximize: bool
    ) -> SpeakerMap:
        num_src, num_tgt = shape
        objective = MaximizationObjective if maximize else MinimizationObjective
        return objective().hard_speaker_map(num_src, num_tgt, assignments)

    @staticmethod
    def correlation(scores1: np.ndarray, scores2: np.ndarray) -> SpeakerMap:
        score_matrix_per_frame = (
            np.stack(  # (local_speakers, num_frames, global_speakers)
                [
                    scores1[:, speaker : speaker + 1] * scores2
                    for speaker in range(scores1.shape[1])
                ],
                axis=0,
            )
        )
        # Calculate total speech "activations" per local speaker
        local_speech_scores = np.sum(scores1, axis=0).reshape(-1, 1)
        # Calculate speaker mapping matrix
        # Cost matrix is the correlation divided by sum of local activations
        score_matrix = np.sum(score_matrix_per_frame, axis=1) / local_speech_scores
        # We want to maximize the correlation to calculate optimal speaker alignments
        return SpeakerMap(score_matrix, MaximizationObjective(max_value=1))

    @staticmethod
    def mse(scores1: np.ndarray, scores2: np.ndarray) -> SpeakerMap:
        cost_matrix = np.stack(  # (local_speakers, local_speakers)
            [
                np.square(scores1[:, speaker : speaker + 1] - scores2).mean(axis=0)
                for speaker in range(scores1.shape[1])
            ],
            axis=0,
        )
        # We want to minimize the MSE to calculate optimal speaker alignments
        return SpeakerMap(cost_matrix, MinimizationObjective())

    @staticmethod
    def mae(scores1: np.ndarray, scores2: np.ndarray) -> SpeakerMap:
        cost_matrix = np.stack(  # (local_speakers, local_speakers)
            [
                np.abs(scores1[:, speaker : speaker + 1] - scores2).mean(axis=0)
                for speaker in range(scores1.shape[1])
            ],
            axis=0,
        )
        # We want to minimize the MSE to calculate optimal speaker alignments
        return SpeakerMap(cost_matrix, MinimizationObjective())

    @staticmethod
    def dist(
        embeddings1: np.ndarray, embeddings2: np.ndarray, metric: Text = "cosine"
    ) -> SpeakerMap:
        # We want to minimize the distance to calculate optimal speaker alignments
        dist_matrix = cdist(embeddings1, embeddings2, metric=metric)
        return SpeakerMap(dist_matrix, MinimizationObjective())

    @staticmethod
    def clf_output(predictions: np.ndarray, pad_to: Optional[int] = None) -> SpeakerMap:
        """
        Parameters
        ----------
        predictions : np.ndarray, (num_local_speakers, num_global_speakers)
            Probability outputs of a speaker embedding classifier
        pad_to : int, optional
            Pad num_global_speakers to this value.
            Useful to deal with unknown speakers that may appear in the future.
            Defaults to no padding
        """
        num_locals, num_globals = predictions.shape
        objective = MaximizationObjective(max_value=1)
        if pad_to is not None and num_globals < pad_to:
            padding = np.ones((num_locals, pad_to - num_globals))
            padding = objective.invalid_value * padding
            predictions = np.concatenate([predictions, padding], axis=1)
        return SpeakerMap(predictions, objective)


class SpeakerMap:
    def __init__(self, mapping_matrix: np.ndarray, objective: MappingMatrixObjective):
        self.mapping_matrix = mapping_matrix
        self.objective = objective
        self.num_source_speakers = self.mapping_matrix.shape[0]
        self.num_target_speakers = self.mapping_matrix.shape[1]
        self.mapped_source_speakers = self.objective.mapped_indices(
            self.mapping_matrix, axis=1
        )
        self.mapped_target_speakers = self.objective.mapped_indices(
            self.mapping_matrix, axis=0
        )
        self._opt_assignments: Optional[List[int]] = None

    @property
    def _raw_optimal_assignments(self) -> List[int]:
        if self._opt_assignments is None:
            self._opt_assignments = self.objective.optimal_assignments(
                self.mapping_matrix
            )
        return self._opt_assignments

    @property
    def shape(self) -> Tuple[int, int]:
        return self.mapping_matrix.shape

    def __len__(self):
        return len(self.mapped_source_speakers)

    def __add__(self, other: SpeakerMap) -> SpeakerMap:
        return self.union(other)

    def _strict_check_valid(self, src: int, tgt: int) -> bool:
        return self.mapping_matrix[src, tgt] != self.objective.invalid_value

    def _loose_check_valid(self, src: int, tgt: int) -> bool:
        return self.is_source_speaker_mapped(src)

    def valid_assignments(
        self,
        strict: bool = False,
        as_array: bool = False,
    ) -> Union[Tuple[List[int], List[int]], Tuple[np.ndarray, np.ndarray]]:
        source, target = [], []
        val_type = "strict" if strict else "loose"
        is_valid = getattr(self, f"_{val_type}_check_valid")
        for src, tgt in enumerate(self._raw_optimal_assignments):
            if is_valid(src, tgt):
                source.append(src)
                target.append(tgt)
        if as_array:
            source, target = np.array(source), np.array(target)
        return source, target

    def is_source_speaker_mapped(self, source_speaker: int) -> bool:
        return source_speaker in self.mapped_source_speakers

    def is_target_speaker_mapped(self, target_speaker: int) -> bool:
        return target_speaker in self.mapped_target_speakers

    def set_source_speaker(self, src_speaker, tgt_speaker: int):
        # if not force:
        #     assert not self.is_source_speaker_mapped(src_speaker)
        #     assert not self.is_target_speaker_mapped(tgt_speaker)
        new_cost_matrix = np.copy(self.mapping_matrix)
        new_cost_matrix[src_speaker, tgt_speaker] = self.objective.best_possible_value
        return SpeakerMap(new_cost_matrix, self.objective)

    def unmap_source_speaker(self, src_speaker: int):
        new_cost_matrix = np.copy(self.mapping_matrix)
        new_cost_matrix[src_speaker] = self.objective.invalid_tensor(
            shape=self.num_target_speakers
        )
        return SpeakerMap(new_cost_matrix, self.objective)

    def unmap_threshold(self, threshold: float) -> SpeakerMap:
        def is_invalid(val):
            if self.objective.maximize:
                return val <= threshold
            else:
                return val >= threshold

        return self.unmap_speakers(
            [
                src
                for src, tgt in zip(*self.valid_assignments())
                if is_invalid(self.mapping_matrix[src, tgt])
            ]
        )

    def unmap_speakers(
        self,
        source_speakers: Optional[Union[List[int], np.ndarray]] = None,
        target_speakers: Optional[Union[List[int], np.ndarray]] = None,
    ) -> SpeakerMap:
        # Set invalid_value to disabled speakers.
        # If they happen to be the best mapping for a local speaker,
        # it means that the mapping of the local speaker should be ignored.
        source_speakers = [] if source_speakers is None else source_speakers
        target_speakers = [] if target_speakers is None else target_speakers
        new_cost_matrix = np.copy(self.mapping_matrix)
        for speaker1 in source_speakers:
            new_cost_matrix[speaker1] = self.objective.invalid_tensor(
                shape=self.num_target_speakers
            )
        for speaker2 in target_speakers:
            new_cost_matrix[:, speaker2] = self.objective.invalid_tensor(
                shape=self.num_source_speakers
            )
        return SpeakerMap(new_cost_matrix, self.objective)

    def compose(self, other: SpeakerMap) -> SpeakerMap:
        """Let's say that `self` is a mapping of `source_speakers` to `intermediate_speakers`
        and `other` is a mapping from `intermediate_speakers` to `target_speakers`.

        Compose `self` with `other` to obtain a new mapping from `source_speakers` to `target_speakers`.
        """
        new_cost_matrix = other.objective.invalid_tensor(
            shape=(self.num_source_speakers, other.num_target_speakers)
        )
        for src_speaker, intermediate_speaker in zip(*self.valid_assignments()):
            target_speaker = other.mapping_matrix[intermediate_speaker]
            new_cost_matrix[src_speaker] = target_speaker
        return SpeakerMap(new_cost_matrix, other.objective)

    def union(self, other: SpeakerMap):
        """`self` and `other` are two maps with the same dimensions.
        Return a new hard speaker map containing assignments in both maps.

        An assignment from `other` is ignored if it is in conflict with
        a source or target speaker from `self`.

        WARNING: The resulting map doesn't preserve soft assignments
        because `self` and `other` might have different objectives.

        :param other: SpeakerMap
            Another speaker map
        """
        assert self.shape == other.shape
        best_val = self.objective.best_possible_value
        new_cost_matrix = self.objective.invalid_tensor(self.shape)
        self_src, self_tgt = self.valid_assignments()
        other_src, other_tgt = other.valid_assignments()
        for src in range(self.num_source_speakers):
            if src in self_src:
                # `self` is preserved by default
                tgt = self_tgt[self_src.index(src)]
                new_cost_matrix[src, tgt] = best_val
            elif src in other_src:
                # In order to add an assignment from `other`,
                # the target speaker cannot be in conflict with `self`
                tgt = other_tgt[other_src.index(src)]
                if not self.is_target_speaker_mapped(tgt):
                    new_cost_matrix[src, tgt] = best_val
        return SpeakerMap(new_cost_matrix, self.objective)

    def apply(self, source_scores: np.ndarray) -> np.ndarray:
        """Apply this mapping to a score matrix of source speakers
        to obtain the same scores aligned to target speakers.

        Parameters
        ----------
        source_scores : SlidingWindowFeature, (num_frames, num_source_speakers)
            Source speaker scores per frame.

        Returns
        -------
        projected_scores : SlidingWindowFeature, (num_frames, num_target_speakers)
            Score matrix for target speakers.
        """
        # Map local speaker scores to the most probable global speaker. Unknown scores are set to 0
        num_frames = source_scores.data.shape[0]
        projected_scores = np.zeros((num_frames, self.num_target_speakers))
        for src_speaker, tgt_speaker in zip(*self.valid_assignments()):
            projected_scores[:, tgt_speaker] = source_scores[:, src_speaker]
        return projected_scores



import random
import time
from queue import SimpleQueue
from typing import Tuple, Text, Optional, Iterable

import sounddevice as sd
from einops import rearrange
from pyannote.audio.core.io import Audio, AudioFile
from pyannote.core import SlidingWindowFeature, SlidingWindow
from rx.subject import Subject


class AudioSource:
    """Represents a source of audio that can start streaming via the `stream` property.

    Parameters
    ----------
    uri: Text
        Unique identifier of the audio source.
    sample_rate: int
        Sample rate of the audio source.
    """
    def __init__(self, uri: Text, sample_rate: int):
        self.uri = uri
        self.sample_rate = sample_rate
        self.stream = Subject()

    @property
    def is_regular(self) -> bool:
        """Whether the stream is regular. Defaults to False.
        A regular stream always yields the same amount of samples per event.
        """
        return False

    @property
    def duration(self) -> Optional[float]:
        """The duration of the stream if known. Defaults to None (unknown duration)"""
        return None

    def read(self):
        """Start reading the source and yielding samples through the stream"""
        raise NotImplementedError


class AudioFileReader:
    """Represents a method for reading an audio file.

    Parameters
    ----------
    sample_rate: int
        Sample rate of the audio file.
    """
    def __init__(self, sample_rate: int):
        self.audio = Audio(sample_rate=sample_rate, mono=True)
        self.resolution = 1 / sample_rate

    @property
    def sample_rate(self) -> int:
        return self.audio.sample_rate

    @property
    def is_regular(self) -> bool:
        """Whether the reading is regular. Defaults to False.
        A regular reading method always yields the same amount of samples."""
        return False

    def get_duration(self, file: AudioFile) -> float:
        return self.audio.get_duration(file)

    def iterate(self, file: AudioFile) -> Iterable[SlidingWindowFeature]:
        """Return an iterable over the file's samples"""
        raise NotImplementedError


class RegularAudioFileReader(AudioFileReader):
    """Reads a file always yielding the same number of samples with a given step.

    Parameters
    ----------
    sample_rate: int
        Sample rate of the audio file.
    window_duration: float
        Duration of each chunk of samples (window) in seconds.
    step_duration: float
        Step duration between chunks in seconds.
    """
    def __init__(
        self,
        sample_rate: int,
        window_duration: float,
        step_duration: float,
    ):
        super().__init__(sample_rate)
        self.window_duration = window_duration
        self.step_duration = step_duration
        self.window_samples = int(round(self.window_duration * self.sample_rate))
        self.step_samples = int(round(self.step_duration * self.sample_rate))

    @property
    def is_regular(self) -> bool:
        return True

    def iterate(self, file: AudioFile) -> Iterable[SlidingWindowFeature]:
        waveform, _ = self.audio(file)
        chunks = rearrange(
            waveform.unfold(1, self.window_samples, self.step_samples),
            "channel chunk frame -> chunk channel frame",
        ).numpy()
        for i, chunk in enumerate(chunks):
            w = SlidingWindow(
                start=i * self.step_duration,
                duration=self.resolution,
                step=self.resolution
            )
            yield SlidingWindowFeature(chunk.T, w)


class IrregularAudioFileReader(AudioFileReader):
    """Reads an audio file yielding a different number of non-overlapping samples in each event.
    This class is useful to simulate how a system would work in unreliable reading conditions.

    Parameters
    ----------
    sample_rate: int
        Sample rate of the audio file.
    refresh_rate_range: (float, float)
        Duration range within which to determine the number of samples to yield (in seconds).
    simulate_delay: bool
        Whether to simulate that the samples are being read in real time before they are yielded.
        Defaults to False (no delay).
    """
    def __init__(
        self,
        sample_rate: int,
        refresh_rate_range: Tuple[float, float],
        simulate_delay: bool = False,
    ):
        super().__init__(sample_rate)
        self.start, self.end = refresh_rate_range
        self.delay = simulate_delay

    def iterate(self, file: AudioFile) -> Iterable[SlidingWindowFeature]:
        waveform, _ = self.audio(file)
        total_samples = waveform.shape[1]
        i = 0
        while i < total_samples:
            rnd_duration = random.uniform(self.start, self.end)
            if self.delay:
                time.sleep(rnd_duration)
            num_samples = int(round(rnd_duration * self.sample_rate))
            last_i = i
            i += num_samples
            yield waveform[:, last_i:i]


class FileAudioSource(AudioSource):
    """Represents an audio source tied to a file.

    Parameters
    ----------
    file: AudioFile
        The file to stream.
    uri: Text
        Unique identifier of the audio source.
    reader: AudioFileReader
        Determines how the file will be read.
    """
    def __init__(
        self,
        file: AudioFile,
        uri: Text,
        reader: AudioFileReader
    ):
        super().__init__(uri, reader.sample_rate)
        self.reader = reader
        self._duration = self.reader.get_duration(file)
        self.file = file

    @property
    def is_regular(self) -> bool:
        """The regularity depends on the reader"""
        return self.reader.is_regular

    @property
    def duration(self) -> Optional[float]:
        """The duration of a file is known"""
        return self._duration

    def read(self):
        """Send each chunk of samples through the stream"""
        for waveform in self.reader.iterate(self.file):
            try:
                self.stream.on_next(waveform)
            except Exception as e:
                self.stream.on_error(e)
        self.stream.on_completed()


class MicrophoneAudioSource(AudioSource):
    """Represents an audio source tied to the default microphone available"""

    def __init__(self, sample_rate: int):
        super().__init__("live_recording", sample_rate)
        self.block_size = 1024
        self.mic_stream = sd.InputStream(
            channels=1,
            samplerate=sample_rate,
            latency=0,
            blocksize=self.block_size,
            callback=self._read_callback
        )
        self.queue = SimpleQueue()

    def _read_callback(self, samples, *args):
        self.queue.put_nowait(samples[:, [0]].T)

    def read(self):
        self.mic_stream.start()
        while self.mic_stream:
            try:
                self.stream.on_next(self.queue.get())
            except Exception as e:
                self.stream.on_error(e)
                break
        self.stream.on_completed()



from typing import Optional

import rx
import rx.operators as ops
from pyannote.audio.pipelines.utils import PipelineModel

from . import functional as fn
from . import operators as dops
from . import sources as src


class OnlineSpeakerDiarization:
    def __init__(
        self,
        segmentation: PipelineModel = "pyannote/segmentation",
        embedding: PipelineModel = "pyannote/embedding",
        duration: Optional[float] = None,
        step: float = 0.5,
        latency: Optional[float] = None,
        tau_active: float = 0.6,
        rho_update: float = 0.3,
        delta_new: float = 1,
        gamma: float = 3,
        beta: float = 10,
        max_speakers: int = 20,
    ):
        self.segmentation = fn.FrameWiseModel(segmentation)
        self.duration = duration
        if self.duration is None:
            self.duration = self.segmentation.model.specifications.duration
        self.embedding = fn.ChunkWiseModel(embedding)
        self.step = step
        self.latency = latency
        if self.latency is None:
            self.latency = self.step
        assert self.step <= self.latency <= self.duration, "Invalid latency requested"
        self.tau_active = tau_active
        self.rho_update = rho_update
        self.delta_new = delta_new
        self.gamma = gamma
        self.beta = beta
        self.max_speakers = max_speakers

    @property
    def sample_rate(self) -> int:
        return self.segmentation.model.audio.sample_rate

    def get_end_time(self, source: src.AudioSource) -> Optional[float]:
        if source.duration is not None:
            return source.duration - source.duration % self.step
        return None

    def from_source(self, source: src.AudioSource, output_waveform: bool = True) -> rx.Observable:
        msg = f"Audio source has sample rate {source.sample_rate}, expected {self.sample_rate}"
        assert source.sample_rate == self.sample_rate, msg
        # Regularize the stream to a specific chunk duration and step
        regular_stream = source.stream
        if not source.is_regular:
            regular_stream = source.stream.pipe(
                dops.regularize_stream(self.duration, self.step, source.sample_rate)
            )
        # Branch the stream to calculate chunk segmentation
        segmentation_stream = regular_stream.pipe(
            ops.map(self.segmentation)
        )
        # Join audio and segmentation stream to calculate speaker embeddings
        osp = fn.OverlappedSpeechPenalty(gamma=self.gamma, beta=self.beta)
        embedding_stream = rx.zip(regular_stream, segmentation_stream).pipe(
            ops.starmap(lambda wave, seg: (wave, osp(seg))),
            ops.starmap(self.embedding),
            ops.map(fn.EmbeddingNormalization(norm=1))
        )
        # Join segmentation and embedding streams to update a background clustering model
        #  while regulating latency and binarizing the output
        clustering = fn.OnlineSpeakerClustering(
            self.tau_active, self.rho_update, self.delta_new, "cosine", self.max_speakers
        )
        end_time = self.get_end_time(source)
        aggregation = fn.DelayedAggregation(
            self.step, self.latency, strategy="hamming", stream_end=end_time
        )
        pipeline = rx.zip(segmentation_stream, embedding_stream).pipe(
            ops.starmap(clustering),
            # Buffer 'num_overlapping' sliding chunks with a step of 1 chunk
            dops.buffer_slide(aggregation.num_overlapping_windows),
            # Aggregate overlapping output windows
            ops.map(aggregation),
            # Binarize output
            ops.map(fn.Binarize(source.uri, self.tau_active)),
        )
        # Add corresponding waveform to the output
        if output_waveform:
            window_selector = fn.DelayedAggregation(
                self.step, self.latency, strategy="first", stream_end=end_time
            )
            waveform_stream = regular_stream.pipe(
                dops.buffer_slide(window_selector.num_overlapping_windows),
                ops.map(window_selector),
            )
            return rx.zip(pipeline, waveform_stream)
        # No waveform needed, add None for consistency
        return pipeline.pipe(ops.map(lambda ann: (ann, None)))
